/**
 * Copyright (C) 2011 Angelo Zerr <angelo.zerr@gmail.com> and Pascal Leclercq <pascal.leclercq@gmail.com>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free  of charge, to any person obtaining
 * a  copy  of this  software  and  associated  documentation files  (the
 * "Software"), to  deal in  the Software without  restriction, including
 * without limitation  the rights to  use, copy, modify,  merge, publish,
 * distribute,  sublicense, and/or sell  copies of  the Software,  and to
 * permit persons to whom the Software  is furnished to do so, subject to
 * the following conditions:
 *
 * The  above  copyright  notice  and  this permission  notice  shall  be
 * included in all copies or substantial portions of the Software.
 *
 * THE  SOFTWARE IS  PROVIDED  "AS  IS", WITHOUT  WARRANTY  OF ANY  KIND,
 * EXPRESS OR  IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE  WARRANTIES OF
 * MERCHANTABILITY,    FITNESS    FOR    A   PARTICULAR    PURPOSE    AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE,  ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package fr.opensagres.xdocreport.document.docx.preprocessor;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import fr.opensagres.xdocreport.document.docx.DocXConstants;
import fr.opensagres.xdocreport.document.preprocessor.sax.TransformedBufferedDocumentContentHandler;
import fr.opensagres.xdocreport.template.formatter.FieldsMetadata;
import fr.opensagres.xdocreport.template.formatter.IDocumentFormatter;

/**
 * SAX content handler to generate lazy Freemarker/Velocity loop directive in
 * the table row which contains a list fields.
 */
public class DocXBufferedDocumentContentHandler extends
		TransformedBufferedDocumentContentHandler implements DocXConstants {

	private RBufferedRegion currentRRegion = null;
	private PBufferedRegion currentPRegion = null;
	private FldSimpleBufferedRegion currentFldSimpleRegion = null;
	private boolean instrTextParsing;
	private boolean tParsing = false;

	protected DocXBufferedDocumentContentHandler(FieldsMetadata fieldsMetadata,
			IDocumentFormatter formater) {
		super(fieldsMetadata, formater);
	}

	@Override
	protected boolean isTable(String uri, String localName, String name) {
		return W_NS.equals(uri) && TBL_ELT.equals(localName);
	}

	@Override
	protected boolean isRow(String uri, String localName, String name) {
		return W_NS.equals(uri) && TR_ELT.equals(localName);
	}

	@Override
	public boolean doStartElement(String uri, String localName, String name,
			Attributes attributes) throws SAXException {
		// Transform mergefield name WordML code with just name of name
		// Merge field is represent with w:fldSimple or w:instrText (complex
		// field). See
		// http://www.documentinteropinitiative.org/implnotes/ecma-376/812d4aca-3071-4352-872a-ca21d65ec913.aspx

		if (isP(uri, localName, name)) {
			// w:p element
			currentPRegion = new PBufferedRegion(currentRegion);
			currentRegion.addRegion(currentPRegion);
			currentRegion = currentPRegion;
			return super.doStartElement(uri, localName, name, attributes);
		}

		if (isR(uri, localName, name) && currentFldSimpleRegion == null) {
			// w:r element
			currentRRegion = new RBufferedRegion(currentRegion);
			currentRegion.addRegion(currentRRegion);
			currentRegion = currentRRegion;
			return super.doStartElement(uri, localName, name, attributes);
		}

		if (isFldChar(uri, localName, name) && currentRRegion != null) {
			// w:fdlChar element
			String fldCharType = attributes.getValue(W_NS, FLDCHARTYPE_ATTR);
			currentRRegion.setFldCharType(fldCharType);
			return super.doStartElement(uri, localName, name, attributes);
		}

		if (isInstrText(uri, localName, name) && currentRRegion != null) {
			// w:instrText element
			instrTextParsing = true;
			return super.doStartElement(uri, localName, name, attributes);
		}

		if (isT(uri, localName, name)) {
			// w:t element
			tParsing = true;
			return super.doStartElement(uri, localName, name, attributes);
		}

		if (isFldSimple(uri, localName, name)) {
			// w:fldSimple element
			// start of fldSimple mergefield, add the fieldName of mergefield
			// and ignore element
			String instrText = processRowIfNeeded(attributes.getValue(W_NS,
					INSTR_ATTR));
			currentFldSimpleRegion = new FldSimpleBufferedRegion(currentRegion);
			currentFldSimpleRegion.setInstrText(instrText);
			currentRegion.addRegion(currentFldSimpleRegion);
			boolean addElement = false;
			if (currentFldSimpleRegion.getFieldName() == null) {
				super.doStartElement(uri, localName, name, attributes);
				addElement = true;
			}
			currentRegion = currentFldSimpleRegion;
			return addElement;

		}

		// Another element
		return super.doStartElement(uri, localName, name, attributes);

	}

	@Override
	public void doEndElement(String uri, String localName, String name)
			throws SAXException {

		if (isP(uri, localName, name) && currentPRegion != null) {
			super.doEndElement(uri, localName, name);
			currentPRegion.process();
			currentRegion = currentPRegion.getParent();
			currentPRegion = null;
			return;
		}

		if (isR(uri, localName, name) && currentRRegion != null
				&& currentFldSimpleRegion == null) {
			super.doEndElement(uri, localName, name);
			currentRegion = currentRRegion.getParent();
			currentRRegion = null;
			return;
		}

		if (isInstrText(uri, localName, name) && currentRRegion != null) {
			super.doEndElement(uri, localName, name);
			instrTextParsing = false;
			return;
		}

		if (isT(uri, localName, name)) {
			super.doEndElement(uri, localName, name);
			tParsing = false;
			return;
		}

		if (isFldSimple(uri, localName, name) && currentFldSimpleRegion != null) {
			// it's end of fldSimple and it's Mergefield; ignore the element
			if (currentFldSimpleRegion.getFieldName() == null) {
				super.doEndElement(uri, localName, name);
			}
			currentRegion = currentFldSimpleRegion.getParent();
			currentFldSimpleRegion = null;
			return;
		}
		super.doEndElement(uri, localName, name);
	}

	@Override
	protected void flushCharacters(String characters) {
		if (tParsing && currentFldSimpleRegion != null) {
			// fldSimple mergefield is parsing, replace with field name.
			currentFldSimpleRegion.setTContent(characters);
			resetCharacters();
			return;
		}

		if (currentRRegion != null) {
			if (instrTextParsing) {
				currentRRegion.setInstrText(processRowIfNeeded(characters));
				resetCharacters();
				return;
			} else {
				if (tParsing) {
					currentRRegion.setTContent(characters);
					resetCharacters();
					return;
				}
			}
		}

		super.flushCharacters(characters);
	}

	/**
	 * Returns true if current element is w:fldSimple and false otherwise.
	 * 
	 * @param uri
	 * @param localName
	 * @param name
	 * @return
	 */
	private boolean isFldSimple(String uri, String localName, String name) {
		return (W_NS.equals(uri) && FLDSIMPLE_ELT.equals(localName));
	}

	/**
	 * Returns true if current element is w:fldSimple with MERGEFIELD type and
	 * false otherwise.
	 * 
	 * If element is w:fldSimple :
	 * 
	 * <pre>
	 * <w:fldSimple w:instr=" MERGEFIELD  ${name} ">
	 * 		<w:r w:rsidR="00396432">
	 * 			<w:rPr>
	 * 				<w:noProof/>
	 * 				</w:rPr>
	 * 				<w:t>${name}</w:t>
	 * 		</w:r>
	 * 	</w:fldSimple>
	 * </pre>
	 * 
	 * it is transformed with
	 * 
	 * <pre>
	 * <w:r w:rsidR="00396432">
	 * 			<w:rPr>
	 * 				<w:noProof/>
	 * 				</w:rPr>
	 * 				<w:t>${name}</w:t>
	 * 		</w:r>
	 * </pre>
	 * 
	 * 
	 * @param uri
	 * @param localName
	 * @param name
	 * @return
	 */

	private boolean isP(String uri, String localName, String name) {
		return (W_NS.equals(uri) && P_ELT.equals(localName));
	}

	protected boolean isR(String uri, String localName, String name) {
		return (W_NS.equals(uri) && R_ELT.equals(localName));
	}

	protected boolean isT(String uri, String localName, String name) {
		return (W_NS.equals(uri) && T_ELT.equals(localName));
	}

	protected boolean isFldChar(String uri, String localName, String name) {
		return (W_NS.equals(uri) && FLDCHAR_ELT.equals(localName));
	}

	protected boolean isInstrText(String uri, String localName, String name) {
		return (W_NS.equals(uri) && INSTR_TEXT_ELT.equals(localName));
	}
}
